<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
	"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">

<html xmlns="http://www.w3.org/1999/xhtml">

<head>
<title>[% title %]</title>
<!-- metadata -->
<meta name="generator" content="[% generator.name %]" />
<meta name="version" content="S5 1.1" />
<meta name="presdate" content="" />
<meta name="author" content="" />
<meta name="company" content="" />
<!-- configuration parameters -->
<meta name="defaultView" content="slideshow" />
<meta name="controlVis" content="hidden" />
<!-- style sheet links -->
<link rel="stylesheet" href="[% resource('ui/default/slides.css') %]" type="text/css" media="projection" id="slideProj" />
<link rel="stylesheet" href="[% resource('ui/default/outline.css') %]" type="text/css" media="screen" id="outlineStyle" />
<link rel="stylesheet" href="[% resource('ui/default/print.css') %]" type="text/css" media="print" id="slidePrint" />
<link rel="stylesheet" href="[% resource('ui/default/opera.css') %]" type="text/css" media="projection" id="operaFix" />
<!-- S5 JS -->
<script src="[% resource('ui/default/slides.js') %]" type="text/javascript"></script>
</head>
<body>

<div class="layout">
<div id="controls"><!-- DO NOT EDIT --></div>
<div id="currentSlide"><!-- DO NOT EDIT --></div>
<div id="header"></div>
<div id="footer">
<!--
<h1>[location/date of presentation]</h1>
<h2>[slide show title here]</h2>
-->
</div>

</div>

<div class="presentation">

[% options.empty_lines = "remove" %]

[% FOR page IN pages %]
[% FOR section IN page.headings(1) %]
<div class="slide">
[% section.content %]
</div>
[% END %]
[% END %]

</div>

</body>
</html>
